#include <linux/module.h>
#include <linux/fs.h>
#include <linux/time.h>
#include "ext3_jbd.h"
#include <linux/jbd.h>
#include <linux/highuid.h>
#include <linux/pagemap.h>
#include <linux/quotaops.h>
#include <linux/string.h>
#include <linux/buffer_head.h>
#include <linux/writeback.h>
#include <linux/mpage.h>
#include <linux/uio.h>
#include <linux/bio.h>
#include <linux/fiemap.h>
#include <linux/namei.h>
#include <linux/mutex.h>
#include "xattr.h"
#include "acl.h"

#include "ldfs_filter.h"
#include "ldfs_store.h"
#include "ddfs.h"
#include "ddfs_proc_fs.h"
#include "ext3_fs.h"

#include <linux/proc_fs.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/mutex.h>
#include <asm/uaccess.h>
#include "ext3_fs.h"
#include "ddfs_proc_fs.h"

struct proc_dir_entry * ddfs_proc_root;
//struct proc_dir_entry * ddfs_proc_print;
//struct proc_dir_entry * ddfs_proc_var;

int ddfs_proc_print_create(struct super_block *sb, struct ddfs_proc_print_entry **entry, char *name)
{
	//TODO: Error Check
	*entry = kzalloc(sizeof(struct ddfs_proc_print_entry), GFP_KERNEL);
	
	(*entry)->dirent = create_proc_entry(name, 0644, EXT3_SB(sb)->part_proc_print);
	((*entry)->dirent)->data = *entry;
	((*entry)->dirent)->read_proc = __ddfs_proc_print_read;

	(*entry)->buf_count = 0; 
	
	mutex_init(&((*entry)->mutex));
	mutex_init(&((*entry)->empty_mutex));
	mutex_init(&((*entry)->full_mutex));
	mutex_init(&((*entry)->print_mutex));
	mutex_lock_interruptible(&(*entry)->empty_mutex);
	mutex_lock_interruptible(&(*entry)->full_mutex);

	(*entry)->enable = FALSE;

	return 0;
}

int ddfs_proc_var_create(struct super_block *sb,
						 struct ddfs_proc_var_entry **entry, 
                         char *name, 
			 void *read_var, 
			 size_t read_size, 
			 void *write_var, 
			 size_t write_size,
			 void (*callback)(struct ddfs_proc_var_entry *))
{
	*entry = kzalloc(sizeof(struct ddfs_proc_var_entry), GFP_KERNEL);
	
	(*entry)->dirent = create_proc_entry(name, 0644, EXT3_SB(sb)->part_proc_var);
	((*entry)->dirent)->data = *entry;

	if(read_var){
		mutex_init(&((*entry)->read_mutex));
		((*entry)->dirent)->read_proc = __ddfs_proc_var_read;
		(*entry)->read_var = read_var;
		(*entry)->read_size = read_size;
		(*entry)->read_count = 0;
	}

	if(write_var){
		mutex_init(&((*entry)->write_mutex));
		((*entry)->dirent)->write_proc = __ddfs_proc_var_write;
		(*entry)->write_var = write_var;
		(*entry)->write_size = write_size;
		(*entry)->callback = callback;
	}

	return 0;
}

int __ddfs_proc_print_read(char *page, char **start, off_t off, int count, int *eof, void *data)
{
	unsigned int byte_write = 0;
	struct ddfs_proc_print_entry *entry = data;

	*start = page;
	*eof = 1;

	entry->enable = TRUE;
	
	mutex_lock_interruptible(&(entry->empty_mutex));

	mutex_lock_interruptible(&(entry->mutex));

	if(entry->buf_count == 0){
		entry->enable = FALSE;
		mutex_unlock(&(entry->mutex));
		return 0;
	}

	if(count >= entry->buf_count){
		byte_write = entry->buf_count;
	}else{
		byte_write = count;
	}

	memcpy(*start, entry->print_buf, byte_write);
	memmove(entry->print_buf, &((entry->print_buf)[byte_write]), sizeof(entry->print_buf) - byte_write);
	
	entry->buf_count -= byte_write;

	if(entry->buf_count){
		mutex_trylock(&(entry->empty_mutex));
		mutex_unlock(&(entry->empty_mutex));
	}else{
		mutex_trylock(&(entry->empty_mutex));
	}
	
	mutex_trylock(&(entry->full_mutex));
	mutex_unlock(&(entry->full_mutex));
	
	mutex_unlock(&(entry->mutex));

	return (byte_write);
}

int __ddfs_proc_var_read(char *page, char **start, off_t off, int count, int *eof, void *data)
{
	unsigned int byte_write = 0;
	unsigned int offset = 0;
	struct ddfs_proc_var_entry *entry = (struct ddfs_proc_var_entry *) data;
	
	*start = page;
	*eof = 1;
	
	if (entry->read_count == 0){
		// First Call Back
		mutex_lock_interruptible(&(entry->read_mutex));
		entry->read_count = entry->read_size;
	}else if(entry->read_count == -1){
		// Last Call Back
		entry->read_count = 0;
		byte_write = 0;
		mutex_unlock(&(entry->read_mutex));
		goto out;
	}
	
	offset = entry->read_size - entry->read_count;

	if(count >= entry->read_count){
		byte_write = entry->read_count;
		entry->read_count = -1;
	}else{
		byte_write = count;
		entry->read_count -= byte_write;
	}

	memcpy(*start, entry->read_var + offset, byte_write);

out:
	return byte_write;
}

int __ddfs_proc_var_write(struct file *file, const char *buffer, unsigned long count, void *data)
{
	int len;
	struct ddfs_proc_var_entry *entry = (struct ddfs_proc_var_entry *) data;

	if(count > entry->write_size){
		len = entry->write_size;
	}else{
		len = count;
	}

	mutex_lock_interruptible(&(entry->write_mutex));

	if(copy_from_user(entry->write_var, buffer, len)){
		return -EFAULT;
	}
	
	mutex_unlock(&(entry->write_mutex));

	if(entry->callback){
		entry->callback(entry);
	}

	return len;
}

int ddfs_proc_readvar(struct ddfs_proc_var_entry *entry, void *var)
{
	mutex_lock_interruptible(&(entry->write_mutex));
	
	memcpy(var, entry->write_var, entry->write_size);

	mutex_unlock(&(entry->write_mutex));

	return 0;
}

int ddfs_proc_writevar(struct ddfs_proc_var_entry *entry, void *var)
{
	mutex_lock_interruptible(&(entry->read_mutex));
	
	memcpy(entry->read_var, var, entry->read_size);

	mutex_unlock(&(entry->read_mutex));

	return 0;
}

int _ddfs_proc_printf(struct ddfs_proc_print_entry *entry, const char *fmt, ...)
{
	va_list args;
    int print_count = 0;
	int reprint = 0;

	if(!entry->enable){
		return -1;
	}

	mutex_lock_interruptible(&(entry->print_mutex));

	mutex_lock_interruptible(&(entry->mutex));
    
	va_start(args, fmt); 
	print_count = vsnprintf(entry->print_tmp, DDFS_PROC_PRINT_MAX, fmt, args);
	if(print_count > DDFS_PROC_PRINT_MAX){
		mutex_unlock(&(entry->mutex));
    	va_end(args);
		return -1;
	}
    va_end(args);

	while(DDFS_PROC_PRINT_BUFSIZE < (entry->buf_count + print_count)){
			mutex_unlock(&(entry->mutex));
			//printk(KERN_EMERG "PROC PRINT: NOT ENOUGH BUFFER SPACE FOR %s\n", (entry->dirent)->name);
			
			mutex_lock_interruptible(&(entry->full_mutex));
			mutex_lock_interruptible(&(entry->mutex));
			
			reprint = 1;
	}
	
	if(reprint){
		vsnprintf(entry->print_tmp, DDFS_PROC_PRINT_MAX, fmt, args);
	}

	strncpy(&(entry->print_buf[entry->buf_count]), entry->print_tmp, print_count);
	entry->buf_count += print_count;

	if(print_count){
		mutex_trylock(&(entry->empty_mutex));
		mutex_unlock(&(entry->empty_mutex));
	}

	mutex_unlock(&(entry->mutex));

	mutex_unlock(&(entry->print_mutex));

	return 0;
}

int ddfs_proc_stream(struct ddfs_proc_print_entry *entry, void *buf, size_t size)
{
    int buf_count = size;

	if(!entry->enable){
		return -1;
	}

	mutex_lock_interruptible(&(entry->print_mutex));
	mutex_lock_interruptible(&(entry->mutex));
    
	if(size > DDFS_PROC_PRINT_MAX){
		mutex_unlock(&(entry->mutex));
		return -1;
	}

	while(DDFS_PROC_PRINT_BUFSIZE < (entry->buf_count + buf_count)){
			mutex_unlock(&(entry->mutex));
			//printk(KERN_EMERG "PROC PRINT: NOT ENOUGH BUFFER SPACE FOR %s\n", (entry->dirent)->name);
			
			mutex_lock_interruptible(&(entry->full_mutex));
			mutex_lock_interruptible(&(entry->mutex));
	}
	
	memcpy(&(entry->print_buf[entry->buf_count]), buf, size);
	entry->buf_count += size;

	if(buf_count){
		mutex_trylock(&(entry->empty_mutex));
		mutex_unlock(&(entry->empty_mutex));
	}

	mutex_unlock(&(entry->mutex));
	mutex_unlock(&(entry->print_mutex));

	return 0;
}

int ddfs_proc_init(void)
{
	// Create a directory under /proc solely for DDFS
	ddfs_proc_root = proc_mkdir("ddfs", NULL);

	return 0;
}

int ddfs_proc_add_dev(struct super_block *sb)
{
	char dev_string[20];

	snprintf(dev_string, 20, "%x", sb->s_dev);

	// Create a device directory
	EXT3_SB(sb)->part_proc_root = proc_mkdir(dev_string, ddfs_proc_root);
	
	// Create a directory under device directory for messages
	EXT3_SB(sb)->part_proc_print = proc_mkdir("msg", EXT3_SB(sb)->part_proc_root);

	// Create a directory under device directory for variables
	EXT3_SB(sb)->part_proc_var = proc_mkdir("var", EXT3_SB(sb)->part_proc_root);
	return 0;
}

int ddfs_proc_remove_dev(struct super_block *sb)
{
	return ddfs_proc_remove_entry(EXT3_SB(sb)->part_proc_root);
}

int ddfs_proc_exit(void)
{
	return ddfs_proc_remove_entry(ddfs_proc_root);
}

int ddfs_proc_remove_entry(struct proc_dir_entry * entry)
{
	struct proc_dir_entry * tmp;

	if(!entry){
		return -1;
	}

	for(tmp = entry->subdir; tmp; tmp = entry->subdir){
		ddfs_proc_remove_entry(tmp);		
	}

	kfree(entry->data);

	remove_proc_entry(entry->name, entry->parent);

	return 0;
}

