#ifndef _DDFS_PROC_FS_H
#define _DDFS_PROC_FS_H

#include <linux/proc_fs.h>
#include <linux/mutex.h>

#define DDFS_PROC_PRINT_MAX 4096
#define DDFS_PROC_PRINT_BUFSIZE 102400

enum{
	FALSE = 0,
	TRUE  = 1
};

struct ddfs_proc_print_entry{
	struct proc_dir_entry *dirent;
	struct mutex mutex;
	struct mutex empty_mutex;
	struct mutex full_mutex;
	struct mutex print_mutex;
	char print_buf[DDFS_PROC_PRINT_BUFSIZE];
	char print_tmp[DDFS_PROC_PRINT_MAX];
	unsigned int buf_count;
	int enable;
	int should_go;
};

struct ddfs_proc_var_entry{
	struct proc_dir_entry *dirent;
	struct mutex read_mutex;
	struct mutex write_mutex;
	void *read_var;
	void *write_var;
	size_t read_size;
	size_t write_size;
	void (*callback)(struct ddfs_proc_var_entry *);
	int read_count;
};

int ddfs_proc_var_create(struct super_block *sb,
						 struct ddfs_proc_var_entry **entry, 
                         char *name, 
			 void *read_var, 
			 size_t read_size, 
			 void *write_var, 
			 size_t write_size,
			 void (*callback)(struct ddfs_proc_var_entry *));
int ddfs_proc_writevar(struct ddfs_proc_var_entry *entry, void *var);

int ddfs_proc_readvar(struct ddfs_proc_var_entry *entry, void *var);

int ddfs_proc_print_create(struct super_block *sb, struct ddfs_proc_print_entry **entry, char *name);
int __ddfs_proc_print_read(char *page, char **start, off_t off, int count, int *eof, void *data);
int __ddfs_proc_var_read(char *page, char **start, off_t off, int count, int *eof, void *data);
int __ddfs_proc_var_write(struct file *file, const char *buffer, unsigned long count, void *data);
//int ddfs_proc_printf(struct ddfs_proc_print_entry *entry, const char *fmt, ...);
#define ddfs_proc_printf(entry, fmt, ...) 
int ddfs_proc_init(void);
int ddfs_proc_exit(void);
int ddfs_proc_stream(struct ddfs_proc_print_entry *entry, void *buf, size_t size);

int ddfs_proc_remove_entry(struct proc_dir_entry * entry);

int ddfs_proc_add_dev(struct super_block *sb);
int ddfs_proc_remove_dev(struct super_block *sb);

extern struct proc_dir_entry * ddfs_proc_root;
extern struct ddfs_proc_print_entry * b_proc_entry ;

#endif //_DDFS_PROC_FS_H
