/*
 *  linux/include/linux/ext3_fs_sb.h
 *
 * Copyright (C) 1992, 1993, 1994, 1995
 * Remy Card (card@masi.ibp.fr)
 * Laboratoire MASI - Institut Blaise Pascal
 * Universite Pierre et Marie Curie (Paris VI)
 *
 *  from
 *
 *  linux/include/linux/minix_fs_sb.h
 *
 *  Copyright (C) 1991, 1992  Linus Torvalds
 */

#ifndef _LINUX_EXT3_FS_SB
#define _LINUX_EXT3_FS_SB

#ifdef __KERNEL__
#include <linux/timer.h>
#include <linux/wait.h>
#include <linux/blockgroup_lock.h>
#include <linux/percpu_counter.h>
#endif
#include <linux/rbtree.h>
#include <linux/list.h>

#include <linux/mutex.h>

//Add By Ming
#include <linux/spinlock_types.h>
#include <linux/spinlock.h>
#include <linux/rbtree.h>
#define DDFS_BARRIER_PAGE_MAX (1024 << 4)
#define DDFS_READ_CACHE_MAX 256
#define	blk_t unsigned
#define DDFS_HASH_LEN 16

struct ldfs_read_cache_page{
    struct rb_node node;
    struct page * page;
    struct list_head lru;
};


//End

//Add By Calix
#include "ddfs_proc_fs.h"


/*
 * third extended-fs super-block data in memory
 */
struct ext3_sb_info {
	unsigned long s_frag_size;	/* Size of a fragment in bytes */
	unsigned long s_frags_per_block;/* Number of fragments per block */
	unsigned long s_inodes_per_block;/* Number of inodes per block */
	unsigned long s_frags_per_group;/* Number of fragments in a group */
	unsigned long s_blocks_per_group;/* Number of blocks in a group */
	unsigned long s_inodes_per_group;/* Number of inodes in a group */
	unsigned long s_itb_per_group;	/* Number of inode table blocks per group */
	unsigned long s_gdb_count;	/* Number of group descriptor blocks */
	unsigned long s_desc_per_block;	/* Number of group descriptors per block */
	unsigned long s_groups_count;	/* Number of groups in the fs */
	unsigned long s_overhead_last;  /* Last calculated overhead */
	unsigned long s_blocks_last;    /* Last seen block count */
	struct buffer_head * s_sbh;	/* Buffer containing the super block */
	struct ext3_super_block * s_es;	/* Pointer to the super block in the buffer */
	struct buffer_head ** s_group_desc;
	unsigned long  s_mount_opt;
	ext3_fsblk_t s_sb_block;
	uid_t s_resuid;
	gid_t s_resgid;
	unsigned short s_mount_state;
	unsigned short s_pad;
	int s_addr_per_block_bits;
	int s_desc_per_block_bits;
	int s_inode_size;
	int s_first_ino;
	spinlock_t s_next_gen_lock;
	u32 s_next_generation;
	u32 s_hash_seed[4];
	int s_def_hash_version;
	int s_hash_unsigned;	/* 3 if hash should be signed, 0 if not */
	struct percpu_counter s_freeblocks_counter;
	struct percpu_counter s_freeinodes_counter;
	struct percpu_counter s_dirs_counter;
	struct blockgroup_lock *s_blockgroup_lock;

	/* root of the per fs reservation window tree */
	spinlock_t s_rsv_window_lock;
	struct rb_root s_rsv_window_root;
	struct ext3_reserve_window_node s_rsv_window_head;

	/* Journaling */
	struct inode * s_journal_inode;
	struct journal_s * s_journal;
	struct list_head s_orphan;
	unsigned long s_commit_interval;
	struct block_device *journal_bdev;
#ifdef CONFIG_JBD_DEBUG
	struct timer_list turn_ro_timer;	/* For turning read-only (crash simulation) */
	wait_queue_head_t ro_wait_queue;	/* For people waiting for the fs to go read-only */
#endif
#ifdef CONFIG_QUOTA
	char *s_qf_names[MAXQUOTAS];		/* Names of quota files with journalled quota */
	int s_jquota_fmt;			/* Format of quota to use */
#endif

	/* LDFS Variables */
	struct kmem_cache * s_bucket_cache;
	struct s_list_head ** s_index_head;
	unsigned s_index_bit;
	struct mutex s_job_mutex;
	struct mutex * s_fingerprint_mutex;

	// Joblist variables
	/*
	struct rb_root s_job_add_hash_root;
	struct rb_root s_job_add_blk_no_root;
	struct rb_root s_job_del_root;
	unsigned s_job_entry_cnt;
	unsigned s_job_entry_max;
	unsigned s_job_state;
	unsigned s_job_flush_delay;

	struct workqueue_struct * s_job_wq;
	struct ddfs_delayed_work * s_job_dw;
	struct kmem_cache * s_job_add_cache;
	struct kmem_cache * s_job_del_cache;
	*/

	//Memory barrier 
	struct inode * barrier_inode;
	unsigned s_next_page;
	unsigned s_page_count;
	unsigned s_max_page;
	spinlock_t s_barrier_page_list_lock;
	unsigned short s_barrier_free_page[DDFS_BARRIER_PAGE_MAX];

    //Read cache
    struct list_head s_read_cache_list;
    struct rb_root s_read_cache_rb_tree;
    struct ldfs_read_cache_page s_cache_pages[DDFS_READ_CACHE_MAX];
	struct mutex s_read_cache_mutex;
	
	struct proc_dir_entry * part_proc_root;
	struct proc_dir_entry * part_proc_print;
	struct proc_dir_entry * part_proc_var;
};

static inline spinlock_t *
sb_bgl_lock(struct ext3_sb_info *sbi, unsigned int block_group)
{
	return bgl_lock_ptr(sbi->s_blockgroup_lock, block_group);
}

#endif	/* _LINUX_EXT3_FS_SB */
