
#ifndef _LIVEDFS_FILTER
#define _LIVEDFS_FILTER
#include <linux/module.h>
#include <linux/string.h>
#include <linux/fs.h>
#include <linux/jbd.h>
#include <linux/slab.h>
#include "ext3_fs.h"
#include "ext3_jbd.h"

#define LDFS_FILTER_BUCKET_MAX 512
#define LDFS_BUCKET_MAX 84
#define LDFS_LOCK_MAX 256

struct s_list_head {
	unsigned long data;		// Two uses, larger 7 bytes is used as pointer, small 1 byte is used as count
};

struct ldfs_bucket_entry {
	char b_key[3];
	char b_storenr[3];
};

struct ldfs_bucket {
	struct s_list_head ptr;
	struct ldfs_bucket_entry entry[LDFS_BUCKET_MAX];
};

struct ldfs_f_bucket {
	__u32 b_key[LDFS_FILTER_BUCKET_MAX];
	__u16 b_blk_no[LDFS_FILTER_BUCKET_MAX];
};

int ldfs_filter_init(struct super_block * sb);
int ldfs_filter_exit(struct super_block * sb);
int ldfs_filter_search(struct super_block * sb, __u8 hash[DDFS_HASH_LEN], __u32 * blk_no);
int ldfs_filter_insert_entry(struct super_block * sb, __u8 hash[DDFS_HASH_LEN], __u32 blk_no);
int ldfs_filter_remove_entry(struct super_block * sb, __u8 hash[DDFS_HASH_LEN], __u32 blk_no);

int ldfs_fingerprint_search(struct super_block * sb, __u8 hash[DDFS_HASH_LEN], __u32 * blk_no);
int ldfs_fingerprint_insert(handle_t * handle, struct super_block * sb, __u8 hash[DDFS_HASH_LEN], __u32 blk_no, __u32 new);
int ldfs_fingerprint_remove(struct super_block * sb, __u8 hash[DDFS_HASH_LEN], __u32 blk_no);

int ldfs_search(struct super_block * sb, __u8 hash[DDFS_HASH_LEN], __u32 * blk_no);
#endif /* _LIVEDFS_FILTER */
